import pandas
import platform
import pythonnet
if (platform.system() == "Linux"):
	pythonnet.load("mono")
if (platform.system() == "Windows"):
	pythonnet.load("netfx")
import clr
clr.AddReference("System.Data")
from System.Data import DataTable

class IStoredProcedures:
	"""
	Interface class for all classes that wish to represent themselves as IStoredProcedure compliant classes.
	"""

	def Execute(self, schema:str, name:str, parameters:list[any] = None) -> None:
		"""
		Execute
		Executes a procedure

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the procedure.

		paramters : list
			Collection of parameters to pass to the stored procedure.
		"""
		raise NotImplementedError()

	def ExecuteWithScalar(self, schema:str, name:str, parameters:list[any] = None) -> any:
		"""
		ExecuteWithScalar
		Executes a procedure and returns the value of first field in the first row.

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the procedure.

		paramters : list
			Collection of parameters to pass to the stored procedure.
		"""
		raise NotImplementedError()

	def ExecuteJSON(self,
			schema:str,
			name:str,
			inputValue:str = None
		) -> str:
		"""
		ExecuteJSON
		Executes a JSON styled procedure which expects an input parameter named Input and an output parameter named Output.

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the procedure.

		inputParamterName : str
			The name of the input paramter for the procedure.

		inputValue : str
			The json string to pass as input.

		outputParamterName : str
			The name of the output paramter for the procedure.
		"""
		raise NotImplementedError()

	def ExecuteJSONInputOnly(self,
				 schema:str,
				 name:str,
				 inputValue:str = None,
		) -> None:
		"""
		ExecuteJSON
		Executes a JSON styled procedure which expects only an input parameter named Input and will have no output.

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the procedure.

		inputValue : str
			The json string to pass as input.
		"""
		raise NotImplementedError()

	def ExecuteJSONOutputOnly(self,
				 schema:str,
				 name:str
		) -> str:
		"""
		ExecuteJSON
		Executes a JSON styled procedure which expects only an output parameter named Output and will not have any inputs.

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the procedure.
		"""
		raise NotImplementedError()

	def ExecuteDataTable(self, schema:str, name:str, parameters:list[any] = None) -> DataTable:
		"""
		ExecuteDataTable
		Executes a procedure and returns the results as a System.Data.DataTable.

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the procedure.

		paramters : list
			Collection of parameters to pass to the stored procedure.
		"""
		raise NotImplementedError()

	def ExecuteDataFrame(self, schema:str, name:str, parameters:list[any] = None) -> pandas.DataFrame:
		"""
		ExecuteDataFrame
		Executes a procedure and returns the results as a pandas.DataFrame.

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the procedure.

		paramters : list
			Collection of parameters to pass to the stored procedure.
		"""
		raise NotImplementedError()

class IQueries:
	"""
	Interface class for all classes that wish to represent themselves as IQueries compliant classes.
	"""

	def Execute(self, query:str, parameters = None):
		"""
		Execute
		Executes a query

		Parameters
		----------
		query : str
			The query to execute.

		paramters : list
			Collection of parameters to pass to the stored procedure.
		"""
		raise NotImplementedError()

	def ExecuteWithScalar(self, query:str, parameters = None):
		"""
		ExecuteWithScalar
		Executes a query and returns the value of first field in the first row.

		Parameters
		----------
		query : str
			The query to execute.

		paramters : list
			Collection of parameters to pass to the stored procedure.
		"""
		raise NotImplementedError()

	def TruncateTable(self, schema:str, name:str):
		"""
		TruncateTable
		Truncates a table.

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the procedure.
		"""
		raise NotImplementedError()

	def GetRowCount(self, schema:str, name:str) -> int:
		"""
		GetRowCount
		Returnes the row count of a table.

		Parameters
		----------
		schema : str
			Name of the schema in which the procedure lives.
		
		name : str
			Name of the table or view.
		"""
		raise NotImplementedError()

class IBulkData:
	def GetDataTable(self, schema:str = None, tableOrView:str = None, query:str = None, options:dict = None) -> DataTable:
		"""
		GetDataTable
		Returns a System.Data.DataTable from either a Schema and Table or from a query.

		Parameters
		----------
		schema : str
			Name of the schema in which the table or view lives.
		
		tableOrView : str
			Name of the table or view.

		query : str
			The query to execute.

		config : dict
			A dictionary of database system specific options.
		"""
		raise NotImplementedError()

	def WriteDataTable(self, schema:str, table:str, dataTable:DataTable, options:dict = None):
		"""
		WriteDataTable
		Writes the contents of a System.Data.DataTable to a table.

		Parameters
		----------
		schema : str
			Name of the schema in which the table or view lives.
		
		table : str
			Name of the table.

		dataTable : DataTable
			The System.Data.DataTable to write.

		config : dict
			A dictionary of database system specific options.
		"""
		raise NotImplementedError()

class IClearData:
	"""
	Interface class for all classes that wish to represent themselves as ICleanData compliant classes.
	"""

	def TestConnection(self) -> dict:
		"""
		TestConnection
		Makes a connection and executes a query that will return some data to prove the connection was successful.
		"""
		raise NotImplementedError()

__all__ = ["IStoredProcedures", "IQueries", "IBulkData", "IClearData"]


from .Interfaces import *
from enum import Enum
from pathlib import Path
import platform
import stat
import datetime
import pandas
import clr
import csv
clr.AddReference('System.Data')
import System
from System.Data import DataTable
from System.Data import DataRow
from System.Data import DataColumn


class FixedWidthJustification(Enum):
	Left = 1
	Right = 2

class TextFileFormat(Enum):
	Delimited = 1
	Prettify = 2
	FixedWidth = 3

class TextFileQuoting(Enum):
	Minimal = 0
	All = 1
	NonNumberic = 2
	NoQuotes = 3
	Strings = 4
	NotNull = 5

	def ToCSVConstant(self) -> int|None:
		returnValue:int|None = None
		match (self):
			case TextFileQuoting.Minimal: returnValue = csv.QUOTE_MINIMAL
			case TextFileQuoting.All: returnValue = csv.QUOTE_ALL
			case TextFileQuoting.NonNumberic: returnValue = csv.QUOTE_NONNUMERIC
			case TextFileQuoting.NoQuotes: returnValue = csv.QUOTE_NONE
			case TextFileQuoting.Strings: returnValue = csv.QUOTE_STRINGS
			case TextFileQuoting.NotNull: returnValue = csv.QUOTE_NOTNULL
			case _: raise ValueError("Unknown")
		return returnValue

class DelimitedFileConfig:
	Quoting:TextFileQuoting|None = None
	FieldDelimiter:str = ","
	RowDelimiter:str = "\n"
	HasHeader:bool = False,
	QuoteCharacter:str = '"'

	def __init__(self, quoting:TextFileQuoting, fieldDelimiter:str=",", rowDelimiter:str="\n", hasHeader:bool=False):
		self.Quoting = quoting
		self.FieldDelimiter = fieldDelimiter
		self.RowDelimiter = rowDelimiter
		self.HasHeader = hasHeader

class FixedWidthFileColumn:
	Ordinal:int|None = None
	Name:str|None=None
	BeginPosition:int|None = None
	EndPosition:int|None = None
	Justification:FixedWidthJustification = FixedWidthJustification.Left
	JustificationCharacter:str|None = " "
	Length:int = 0
	TrimJustificationCharacters:bool = False

	def __init__(self,
					ordinal:int|None=None, name:int|None=None,
					beginPosition:int|None=None, endPosition:int|None=None,
					justification:FixedWidthJustification = FixedWidthJustification.Left,
					justificationCharacter:str|None=" ",
					trimJustificationCharacters:bool=False):
		self.Ordinal = ordinal
		self.Name = name
		self.BeginPosition = beginPosition
		self.EndPosition = endPosition
		self.Justification = justification
		self.JustificationCharacter = justificationCharacter
		self.TrimJustificationCharacters = trimJustificationCharacters
		self.Length = (self.EndPosition - self.BeginPosition) + 1

class FixedWidthFileColumns(list):
	LastUsedOrdinal:int = 0

	def __len__(self):
		return super().__len__()

	def __iter__(self):
		return super().__iter__()

	def __getitem__(self, item):
		return super()[item]

	def __init__(self) -> None:
		super().__init__()

	def append(self, object:FixedWidthFileColumn):
		if (object.Ordinal is None):
			self.LastUsedOrdinal += 1
			object.Ordinal = self.LastUsedOrdinal
		elif (object.Ordinal == 0):
			self.LastUsedOrdinal += 1
			object.Ordinal = self.LastUsedOrdinal
		super().append(object)

	def append(self, object:FixedWidthFileColumn):
		if (object.Ordinal is None):
			self.LastUsedOrdinal += 1
			object.Ordinal = self.LastUsedOrdinal
		elif (object.Ordinal == 0):
			self.LastUsedOrdinal += 1
			object.Ordinal = self.LastUsedOrdinal
		super().append(object)

	def GetColumn(name:str) -> FixedWidthFileColumn|None:
		return next(filter(lambda x: x.Name == name,  super()), None)

class TextQueries(IQueries):
	FilePath:Path = None
	Format:TextFileFormat = TextFileFormat.Delimited
	Columns:FixedWidthFileColumns|None = None
	DelimitedConfig:DelimitedFileConfig|None = None

	def __init__(self, filePath:Path, format:TextFileFormat = TextFileFormat.Delimited, delimitedConfig:DelimitedFileConfig|None = None, fixedWidthColumns:FixedWidthFileColumns|None = None):
		self.FilePath = filePath
		self.Format = format
		self.Columns = fixedWidthColumns
		self.DelimitedConfig = delimitedConfig

	def Execute(self, query:str, parameters = None):
		raise NotImplementedError()

	def ExecuteWithScalar(self, query:str, parameters = None):
		raise NotImplementedError()

	def ExecuteDataTable(self, query:str, parameters = None):
		raise NotImplementedError()

	def ExecuteDataFrame(self, query:str, parameters = None):
		raise NotImplementedError()

	def TruncateTable(self):
		header:str = None
		if (self.Format == TextFileFormat.Delimited
			and self.DelimitedConfig.HasHeader):
			header = self.FilePath.read_text().split(self.DelimitedConfig.RowDelimiter)[0]
		self.FilePath.write_text(header)

	def GetRowCount(self) -> int:
		returnValue:int = None
		if (self.Format != TextFileFormat.Prettify):
			contents:str = self.FilePath.read_text()
			returnValue = len(contents.split("\n"))
			returnValue -= 1
			match self.Format:
				case TextFileFormat.CommaWithHeader:
					returnValue -= 1
				case TextFileFormat.TabWithHeader:
					returnValue -= 1
				case TextFileFormat.PipeWithHeader:
					returnValue -= 1
		return returnValue

class TextBulkData:
	FilePath:Path = None
	Format:TextFileFormat = TextFileFormat.Delimited
	FixedWidthColumns:FixedWidthFileColumns|None = None
	DelimitedConfig:DelimitedFileConfig|None = None

	def __init__(self, filePath:Path, format:TextFileFormat = TextFileFormat.Delimited, delimitedConfig:DelimitedFileConfig|None = None, fixedWidthColumns:FixedWidthFileColumns|None = None):
		self.FilePath = filePath
		self.Format = format
		self.FixedWidthColumns = fixedWidthColumns
		self.DelimitedConfig = delimitedConfig

	def __GetDataTableDelimited__(self) -> DataTable:
		returnValue:DataTable = None
		with open(self.FilePath, newline='', encoding="UTF-8") as file:
			returnValue = DataTable()
			reader = csv.reader(file,
						delimiter=self.DelimitedConfig.FieldDelimiter,
						lineterminator=self.DelimitedConfig.RowDelimiter,
						quotechar=self.DelimitedConfig.QuoteCharacter,
						quoting=self.DelimitedConfig.Quoting.ToCSVConstant())
			rowNumber:int = 0
			for row in reader:
				rowNumber += 1
				if (rowNumber == 1):
					columns:list = []
					if (self.DelimitedConfig.HasHeader):
						for item in row:
							columnName:str = item
							dupeIndex:int = 0
							while (columns.count(columnName) > 0):
								dupeIndex += 1
								columnName = f"{columnName}_{dupeIndex}"
							columns.append(columnName)
					else:
						columns = ["Field_{}".format(x, y) for (x, y) in enumerate(row)]
					for columnName in columns:
						dataColumn:DataColumn = DataColumn(columnName)
						dataColumn.DataType = System.Type.GetType("System.String")
						dataColumn.AllowDBNull  = True
						returnValue.Columns.Add(dataColumn)
					if (not self.DelimitedConfig.HasHeader):
						dataRow:DataRow = returnValue.NewRow()
						for index, item in enumerate(row):
							dataRow[index] = item
						returnValue.Rows.Add(dataRow)
				else:
					dataRow:DataRow = returnValue.NewRow()
					for index, item in enumerate(row):
						dataRow[index] = item
					returnValue.Rows.Add(dataRow)
		return returnValue

	def __GetDataTableFixedWidth__(self) -> DataTable:
		returnValue:DataTable = DataTable()
		for fixedWidthColumn in self.FixedWidthColumns:
			dataColumn:DataColumn = DataColumn(fixedWidthColumn.Name)
			dataColumn.DataType = System.Type.GetType("System.String")
			dataColumn.AllowDBNull  = True
			returnValue.Columns.Add(dataColumn)
		for lineIndex, lineText in enumerate(self.FilePath.read_text().splitlines()):
			values:list[str] = list[str]()
			for fixedWidthColumn in self.FixedWidthColumns:
				value:str = lineText[(fixedWidthColumn.BeginPosition-1):fixedWidthColumn.EndPosition]
				if (fixedWidthColumn.TrimJustificationCharacters):
					if (fixedWidthColumn.Justification == FixedWidthJustification.Right):
						value = value.rstrip(fixedWidthColumn.JustificationCharacter)
					elif (fixedWidthColumn.Justification == FixedWidthJustification.Left):
						value = value.lstrip(fixedWidthColumn.JustificationCharacter)
				values.append(value)
			returnValue.Rows.Add(values)
		return returnValue

	def __WriteDataTableDelimited__(self, dataTable:DataTable):
		with open(self.FilePath, newline='', encoding="UTF-8") as file:
			writer = csv.writer(file,
					delimiter=self.DelimitedConfig.FieldDelimiter,
					lineterminator=self.DelimitedConfig.RowDelimiter,
					quotechar=self.DelimitedConfig.QuoteCharacter,
					quoting=self.DelimitedConfig.Quoting.ToCSVConstant())
			if (self.DelimitedConfig.HasHeader):
				header:list[str] = list[str]()
				for dataColumn in dataTable.Columns:
					header.append(dataColumn.ColumnName)
				writer.writerow(header)
			for dataRow in dataTable.Rows:
				values:list[any] = list[any]()
				for dataColumn in dataTable.Columns:
					values.append(dataRow[dataColumn.ColumnName])
				writer.writerow(values)

	def __WriteDataTableFixedWidth__(self, dataTable:DataTable):
		rowDelimiter:str = "\n"
		lines:list = []
		line:str = None
		line = ""
		for dataRow in dataTable.Rows:
			line = ""
			for dataColumn in dataTable.Columns:
				fixedWidthFileColumn:FixedWidthFileColumn|None = self.FixedWidthColumns.GetColumn(dataColumn.ColumnName)
				if (fixedWidthFileColumn is not None):
					if (fixedWidthFileColumn.Justification == FixedWidthJustification.Left):
						line += str(dataRow[dataColumn.ColumnName]).ljust(fixedWidthFileColumn.Length, fixedWidthFileColumn.JustificationCharacter)
					if (fixedWidthFileColumn.Justification == FixedWidthJustification.Right):
						line += str(dataRow[dataColumn.ColumnName]).rjust(fixedWidthFileColumn.Length, fixedWidthFileColumn.JustificationCharacter)
			lines.append(line)
		self.FilePath.write_text(rowDelimiter.join(lines))

	def __WriteDataTablePrettify__(self, dataTable:DataTable):
		returnValue:str = None
		exception = None
		try:
			totalColumnWidth:int = 0
			columnWidth:dict = {}
			lines:list = []
			line:str = None
			allLines:str = ""
			for dataRow in dataTable.Rows:
				for dataColumn in dataTable.Columns:
					if (dataColumn.ColumnName not in columnWidth):
						columnWidth[dataColumn.ColumnName] = len(dataColumn.ColumnName)
					if (columnWidth[dataColumn.ColumnName] < len(str(dataRow[dataColumn]))):
						columnWidth[dataColumn.ColumnName] = len(str(dataRow[dataColumn]))
			for columnName in columnWidth:
				totalColumnWidth += columnWidth[columnName]
			lines.append(" -" + ("-"*totalColumnWidth) + ("--"*len(columnWidth)) + "- ")
			line = ""
			for columnName in columnWidth:
				line += f" {columnName.ljust(columnWidth[columnName])} "
			lines.append(f"| {line} |")
			line = ""
			for columnName in columnWidth:
				headBreak:str = "-"*columnWidth[columnName]
				line += f" {headBreak} "
			lines.append(f"| {line} |")
			for dataRow in dataTable.Rows:
				line = ""
				for dataColumn in dataTable.Columns:
					if (str(dataColumn.DataType) in [
						"System.Byte",
						"System.Int16",
						"System.Int32",
						"System.Int64",
						"System.Decimal"]):
						line += f" {str(dataRow[dataColumn.ColumnName]).rjust(columnWidth[dataColumn.ColumnName])} "
					else:
						line += f" {str(dataRow[dataColumn.ColumnName]).ljust(columnWidth[dataColumn.ColumnName])} "
				lines.append(f"| {line} |")
			lines.append(" -" + ("-"*totalColumnWidth) + ("--"*len(columnWidth)) + "- ")
			returnValue = ""
			for ln in lines:
				returnValue += f"{ln}\n"
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		elif returnValue is not None:
			return returnValue

	def GetDataTable(self) -> DataTable|None:
		returnValue:DataTable|None = None
		match (self.Format):
			case TextFileFormat.Delimited:
				returnValue = self.__GetDataTableDelimited__()
			case TextFileFormat.FixedWidth:
				returnValue = self.__GetDataTableFixedWidth__()
			case _:
				returnValue = None
		return returnValue

	def WriteDataTable(self, dataTable:DataTable):
		match (self.Format):
			case TextFileFormat.Delimited:
				returnValue = self.__WriteDataTableDelimited__(dataTable)
			case TextFileFormat.FixedWidth:
				returnValue = self.__WriteDataTableFixedWidth__(dataTable)
			case _:
				returnValue = None

	def GetDataFrame(self, schema:str = None, tableOrView:str = None, options:dict = None) -> pandas.DataFrame:
		raise NotImplementedError()

	def WriteDataFrame(self, schema:str, table:str, dataFrame:pandas.DataFrame, options:dict = None):
		raise NotImplementedError()

class TextClearData(IClearData):
	FilePath:Path = None
	Format:TextFileFormat = TextFileFormat.Delimited

	@property
	def StoredProcedures(self):
		raise NotImplementedError("TextClearData does not implement IStoredProcedures")

	def __init__(self, filePath:Path, format:TextFileFormat = TextFileFormat.Delimited, delimitedConfig:DelimitedFileConfig|None = None, fixedWidthColumns:FixedWidthFileColumns|None = None):
		if (format == TextFileFormat.Delimited
			and delimitedConfig is None):
			raise ValueError("If format is delimited, then delimitedConfig must be provided")
		if (format == TextFileFormat.FixedWidth
			and fixedWidthColumns is None):
			raise ValueError("If format is fixed wicth, then fixedWidthColumns must be provided")
		self.FilePath = filePath
		self.Format = format
		self.Queries = TextQueries(filePath, format, delimitedConfig, fixedWidthColumns)
		self.BulkData = TextBulkData(filePath, format, delimitedConfig, fixedWidthColumns)

	def TestConnection(self) -> dict:
		returnValue:dict = dict()
		if (self.FilePath.exists()):
			filestat = self.FilePath.stat()
			returnValue.update({
					"Size": filestat.st_size,
					"CreateTime": datetime.datetime.fromtimestamp(filestat.st_birthtime),
					"LastAccessTime": datetime.datetime.fromtimestamp(filestat.st_atime),
					"LastModifiedTime": datetime.datetime.fromtimestamp(filestat.st_mtime),
					"CreateTime": datetime.datetime.fromtimestamp(filestat.st_birthtime)
				})
			if (platform.system() == "Linux"):
				returnValue.update({
					"FileSystem": filestat.st_fstype,
					"Type": filestat.st_type,
					"Creator": filestat.st_creator
				})
			if (platform.system() == "Windows"):
				returnValue.update({ "FileAttributes": filestat.st_file_attributes })
				returnValue.update({ "IsArchive": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_ARCHIVE) == stat.FILE_ATTRIBUTE_ARCHIVE) })
				returnValue.update({ "IsCompressed": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_COMPRESSED) == stat.FILE_ATTRIBUTE_COMPRESSED) })
				returnValue.update({ "IsDecive": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_DEVICE) == stat.FILE_ATTRIBUTE_DEVICE) })
				returnValue.update({ "IsDirectory": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_DIRECTORY) == stat.FILE_ATTRIBUTE_DIRECTORY) })
				returnValue.update({ "IsEncrypted": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_ENCRYPTED) == stat.FILE_ATTRIBUTE_ENCRYPTED) })
				returnValue.update({ "IsHidden": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN) == stat.FILE_ATTRIBUTE_HIDDEN) })
				returnValue.update({ "IsIntegrityStream": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_INTEGRITY_STREAM) == stat.FILE_ATTRIBUTE_INTEGRITY_STREAM) })
				returnValue.update({ "IsNormal": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_NORMAL) == stat.FILE_ATTRIBUTE_NORMAL) })
				returnValue.update({ "IsContentNotIndexed": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_NOT_CONTENT_INDEXED) == stat.FILE_ATTRIBUTE_NOT_CONTENT_INDEXED) })
				returnValue.update({ "IsNoScrubData": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_NO_SCRUB_DATA) == stat.FILE_ATTRIBUTE_NO_SCRUB_DATA) })
				returnValue.update({ "IsOffline": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_OFFLINE) == stat.FILE_ATTRIBUTE_OFFLINE) })
				returnValue.update({ "IsReadOnly": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_READONLY) == stat.FILE_ATTRIBUTE_READONLY) })
				returnValue.update({ "IsReparsePoint": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_REPARSE_POINT) == stat.FILE_ATTRIBUTE_REPARSE_POINT) })
				returnValue.update({ "IsSpareFile": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_SPARSE_FILE) == stat.FILE_ATTRIBUTE_SPARSE_FILE) })
				returnValue.update({ "IsSystem": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_SYSTEM) == stat.FILE_ATTRIBUTE_SYSTEM) })
				returnValue.update({ "IsTemporary": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_TEMPORARY) == stat.FILE_ATTRIBUTE_TEMPORARY) })
				returnValue.update({ "IsVirtual": ((filestat.st_file_attributes & stat.FILE_ATTRIBUTE_VIRTUAL) == stat.FILE_ATTRIBUTE_VIRTUAL) })

	@staticmethod
	def GetTextFileFormatConfig(format:TextFileFormat = TextFileFormat.Delimited):
		fieldDelimiter:str = ","
		rowDelimiter:str = "\n"
		includeHeader:bool = True
		match format:
			case TextFileFormat.CommaWithHeader:
				fieldDelimiter = ","
				rowDelimiter = "\n"
				includeHeader = True
			case TextFileFormat.CommaWithoutHeader:
				fieldDelimiter = ","
				rowDelimiter = "\n"
				includeHeader = False
			case TextFileFormat.TabWithHeader:
				fieldDelimiter = "\t"
				rowDelimiter = "\n"
				includeHeader = True
			case TextFileFormat.TabWithoutHeader:
				fieldDelimiter = "\t"
				rowDelimiter = "\n"
				includeHeader = False
			case TextFileFormat.PipeWithHeader:
				fieldDelimiter = "|"
				rowDelimiter = "\n"
				includeHeader = True
			case TextFileFormat.PipeWithoutHeader:
				fieldDelimiter = "|"
				rowDelimiter = "\n"
				includeHeader = False
			case TextFileFormat.FixedWidth:
				fieldDelimiter = None
				rowDelimiter = "\n"
				includeHeader = False
		return fieldDelimiter, rowDelimiter, includeHeader

	@staticmethod
	def ConvertFixedWidthToDelimited(
			fixedWidthFilePath:Path, fixedWidthColumns:FixedWidthFileColumns,
			delimitedFilePath:Path, delimitedConfig:DelimitedFileConfig):
		fixedBulkData = TextBulkData(filePath=fixedWidthFilePath, format=TextFileFormat.FixedWidth, delimitedConfig=None, fixedWidthColumns=fixedWidthColumns)
		delimitedBulkData = TextBulkData(filePath=delimitedFilePath, format=TextFileFormat.Delimited, delimitedConfig=delimitedConfig, fixedWidthColumns=None)
		delimitedBulkData.WriteDataTable(fixedBulkData.GetDataTable())

	@staticmethod
	def ConvertDelimitedToFixedWidth(
			fixedWidthFilePath:Path, fixedWidthColumns:FixedWidthFileColumns,
			delimitedFilePath:Path, delimitedConfig:DelimitedFileConfig):
		fixedBulkData = TextBulkData(filePath=fixedWidthFilePath, format=TextFileFormat.FixedWidth, delimitedConfig=None, fixedWidthColumns=fixedWidthColumns)
		delimitedBulkData = TextBulkData(filePath=delimitedFilePath, format=TextFileFormat.Delimited, delimitedConfig=delimitedConfig, fixedWidthColumns=None)
		fixedBulkData.WriteDataTable(delimitedBulkData.GetDataTable())

__all__ = ["FixedWidthJustification", "TextFileFormat", "TextFileQuoting",
			"DelimitedFileConfig", "FixedWidthFileColumn", "FixedWidthFileColumns",
			"TextQueries", "TextBulkData", "TextClearData"]



from .Interfaces import *
import platform
import pythonnet
if (platform.system() == "Linux"):
	pythonnet.load("mono")
if (platform.system() == "Windows"):
	pythonnet.load("netfx")
import clr
clr.AddReference("System")
clr.AddReference("System.Data")
from System.Data import DataTable
from System.Data import DataRow
from System.Data import DataColumn
from System.Data import ParameterDirection

import pandas
import psycopg
from psycopg.rows import DictRow, dict_row
import datetime
import json
from pathlib import Path

class PostgreSQLParameter:
	def __init__(self, name:str, dataType:str, direction:ParameterDirection, value, size:int = None):
		self.Name = name
		self.Direction = direction
		self.Value = value

class PostgreSQLStoredProcedures(IStoredProcedures):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def Execute(self, schema:str, name:str, parameters:dict = None) -> None:
		exception = None
		try:
			parameterText:str = None
			if (parameters is None):
				parameterText = ""
			else:
				for key in parameters.keys():
					if (parameterText is None):
						parameterText = f"%({key})s"
					else:
						parameterText += f", %({key})s"
			commandText:str = f"SELECT * FROM \"{schema}\".\"{name}\"({parameterText}) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, parameters)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def ExecuteWithScalar(self, schema:str, name:str, parameters:dict = None) -> any:
		returnValue:any | None = None
		exception = None
		try:
			parameterText:str = None
			if (parameters is None):
				parameterText = ""
			else:
				for key in parameters.keys():
					if (parameterText is None):
						parameterText = f"%({key})s"
					else:
						parameterText += f", %({key})s"
			commandText:str = f"SELECT * FROM \"{schema}\".\"{name}\"({parameterText}) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, parameters)
					for row in cur:
						returnValue = row[0]
						break
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def ExecuteJSON(self,
			schema:str,
			name:str,
			inputValue:str = None
		) -> str:
		returnValue:str | None = None
		exception = None
		try:
			commandText:str = f"SELECT FROM \"{schema}\".\"{name}\"(%(Input)s) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, {"Input": inputValue})
					for row in cur:
						returnValue = row[0]
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def ExecuteJSONInputOnly(self,
				 schema:str,
				 name:str,
				 inputValue:str = None,
		) -> None:
		exception = None
		try:
			commandText:str = f"SELECT FROM \"{schema}\".\"{name}\"(%(Input)s) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, {"Input": inputValue})
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def ExecuteJSONOutputOnly(self,
				 schema:str,
				 name:str
		) -> str:
		returnValue:str | None = None
		exception = None
		try:
			commandText:str = f"SELECT FROM \"{schema}\".\"{name}\"() AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
					for row in cur:
						returnValue = row[0]
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def ExecuteDataTable(self, schema:str, name:str, parameters:dict = None) -> DataTable:
		returnValue = None
		exception = None
		try:
			parameterText:str = None
			if (parameters is None):
				parameterText = ""
			else:
				for key in parameters.keys():
					keyValue:str = "NULL"
					if (parameters[key] is not None): keyValue = f"'{parameters[key]}'"
					keyValue = f"\"{key}\"=>{keyValue}"
					if (parameterText is None):
						parameterText = f"{keyValue}"
					else:
						parameterText += f", {keyValue}"
			commandText:str = f"SELECT * FROM \"{schema}\".\"{name}\"({parameterText}) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, parameters)
					returnValue = DataTable()
					rowCount:int = 0
					for row in cur:
						rowCount += 1
						dataRow = returnValue.NewRow()
						for columnIndex, column in enumerate(cur.description):
							clrDataType:str = PostgreSQLClearData.__GetCLRDataType__(row[columnIndex])
							if (not returnValue.Columns.Contains(column.name)):
								dataColumn:DataColumn = DataColumn(column.name)
								dataColumn.DataType = System.Type.GetType(clrDataType)
								dataColumn.AllowDBNull  = True
								#returnValue.Columns.Add(column.name, System.Type.GetType(clrDataType))
								returnValue.Columns.Add(dataColumn)
							dataRow[columnIndex] = PostgreSQLClearData.__ConvertToCLRDataType__(row[columnIndex], clrDataType)
						returnValue.Rows.Add(dataRow)

		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def ExecuteDataFrame(self, schema:str, name:str, parameters = None) -> pandas.DataFrame:
		returnValue = pandas.DataFrame()
		returnValue = None
		exception = None
		try:
			parameterText:str = None
			if (parameters is None):
				parameterText = ""
			else:
				for key in parameters.keys():
					keyValue:str = "NULL"
					if (parameters[key] is not None): keyValue = f"'{parameters[key]}'"
					keyValue = f"\"{key}\"=>{keyValue}"
					if (parameterText is None):
						parameterText = f"{keyValue}"
					else:
						parameterText += f", {keyValue}"
			commandText:str = f"SELECT * FROM \"{schema}\".\"{name}\"({parameterText}) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, parameters)
					data = cur.fetchall()
					returnValue = pandas.DataFrame(data, columns=[desc[0] for desc in cur.description])
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

class PostgreSQLQueries(IQueries):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def Execute(self, query:str):
		exception = None
		try:
			commandText:str = query
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def ExecuteWithScalar(self, query:str):
		returnValue = None
		exception = None
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(query)
					for row in cur:
						returnValue = row[0]
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

	def ExecuteDataTable(self, query:str) -> DataTable:
		returnValue = None
		exception = None
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(query)
					returnValue = DataTable()
					rowCount:int = 0
					for row in cur:
						rowCount += 1
						dataRow = returnValue.NewRow()
						for columnIndex, column in enumerate(cur.description):
							clrDataType:str = PostgreSQLClearData.__GetCLRDataType__(row[columnIndex])
							if (not returnValue.Columns.Contains(column.name)):
								dataColumn:DataColumn = DataColumn(column.name)
								dataColumn.DataType = System.Type.GetType(clrDataType)
								dataColumn.AllowDBNull  = True
								returnValue.Columns.Add(dataColumn)
							dataRow[columnIndex] = PostgreSQLClearData.__ConvertToCLRDataType__(row[columnIndex], clrDataType)
						returnValue.Rows.Add(dataRow)

		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def ExecuteDataFrame(self, query:str) -> pandas.DataFrame:
		returnValue = pandas.DataFrame()
		returnValue = None
		exception = None
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(query)
					data = cur.fetchall()
					returnValue = pandas.DataFrame(data, columns=[desc[0] for desc in cur.description])
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def TruncateTable(self, schema:str, table:str):
		exception = None
		try:
			commandText:str = f"TRUNCATE TABLE {schema}.{table}"
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def GetRowCount(self, schema:str, table:str) -> int:
		returnValue:int = None
		exception = None
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(f"SELECT COUNT(*) AS \"Count\" FROM \"{schema}\".\"{table}\"")
					returnValue = int(cur.fetchone()[0])
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

class PostgreSQLBulkData(IBulkData):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def GetDataTable(
		self,
		schema:str = None,
		tableOrView:str = None,
		options:dict = {"UseBinaryFormat": False, "CLRTypeOverride": {} }
	):
		returnValue = None
		exception = None
		commandText:str = f"SELECT * FROM \"{schema}\".\"{tableOrView}\""
		useBinaryFormat:bool = False
		clrTypeOverride:dict = {}
		if ("BatchSize" in options):
			useBinaryFormat = bool(options["UseBinaryFormat"])
		if ("CLRTypeOverride" in options):
			if (isinstance(options["CLRTypeOverride"], dict)):
				clrTypeOverride = options["CLRTypeOverride"]
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
					returnValue = DataTable()
					rowCount:int = 0
					for row in cur:
						rowCount += 1
						dataRow = returnValue.NewRow()
						for columnIndex, column in enumerate(cur.description):
							clrDataType:str = None
							if (column.name not in clrTypeOverride):
								clrDataType:str = PostgreSQLClearData.__GetCLRDataType__(row[columnIndex])
							else:
								clrDataType = clrTypeOverride[column.name]
							if (not returnValue.Columns.Contains(column.name)):
								dataColumn:DataColumn = DataColumn(column.name)
								dataColumn.DataType = System.Type.GetType(clrDataType)
								dataColumn.AllowDBNull  = True
								#returnValue.Columns.Add(column.name, System.Type.GetType(clrDataType))
								returnValue.Columns.Add(dataColumn)
							dataRow[columnIndex] = PostgreSQLClearData.__ConvertToCLRDataType__(row[columnIndex], clrDataType)
						returnValue.Rows.Add(dataRow)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		elif returnValue is not None:
			return returnValue

	def WriteDataTable(self, schema:str, table:str, dataTable:DataTable, options:dict = {"UseBinaryFormat": False}):
		exception = None
		binaryFormat:str = ""
		useBinaryFormat:bool = False
		if ("BatchSize" in options):
			useBinaryFormat = bool(options["UseBinaryFormat"])
		if (useBinaryFormat):
			binaryFormat = " (FORMAT BINARY)"
		try:
			columnList:str = None
			for column in dataTable.Columns:
				if (columnList is None):
					columnList = f"\"{column.ColumnName}\""
				else:
					columnList += f", \"{column.ColumnName}\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					returnValue = DataTable()
					with cur.copy(f"COPY \"{schema}\".\"{table}\" ({columnList}) FROM STDOUT{binaryFormat}") as copy:
						for row in dataTable.Rows:
							rowData = []
							for item in row.ItemArray:
								if (isinstance(item, System.DBNull)):
									rowData.append(None)
								elif (isinstance(item, System.DateTime)):
									rowData.append(datetime.datetime.fromisoformat(item.ToString("yyyy-MM-ddTHH:mm:ss.fffffff")))
								else:
									rowData.append(item)
							copy.write_row(rowData)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def GetDataFrame(self, schema:str = None, tableOrView:str = None, options:dict = None) -> pandas.DataFrame:
		raise NotImplementedError()

	def WriteDataFrame(self, schema:str, table:str, dataFrame:pandas.DataFrame, options:dict = None):
		raise NotImplementedError()

class PostgreSQLClearData:
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString
		self.StoredProcedures = PostgreSQLStoredProcedures(connectionString)
		self.Queries = PostgreSQLQueries(connectionString)
		self.BulkData = PostgreSQLBulkData(connectionString)

	def __GetTestConnectionQuery__(self) -> str:
		return """
			SELECT ROW_TO_JSON(\"Sections\")
				FROM
				(
					SELECT
						(
							SELECT ROW_TO_JSON(\"Client\")
								FROM
								(
									SELECT
										CURRENT_SETTING('application_name') AS \"ApplicationName\",
										inet_client_addr() AS \"HostAddress\",
										inet_client_port() AS \"HostPort\",
										CURRENT_ROLE AS \"Role\",
										CURRENT_USER AS \"CurrentUser\",
										USER AS \"User\"
								) AS \"Client\"
						) AS \"Client\",
						(
							SELECT ROW_TO_JSON(\"Server\")
								FROM
								(
									SELECT
										inet_server_addr() AS \"HostAddress\",
										inet_server_port() AS \"HostPort\",
										CURRENT_SETTING('SERVER_VERSION') AS \"Version\"
								) AS \"Server\"
						) AS \"Server\",
						(
							SELECT ROW_TO_JSON(\"Database\")
								FROM
								(
									SELECT
										current_database()  AS \"Name\"
								) AS \"Database\"
						) AS \"Database\"
				) AS \"Sections\"
		"""

	def TestConnection(self) -> dict:
		returnValue:dict = None
		commandText:str = self.__GetTestConnectionQuery__()
		with open(Path(__file__).parent.joinpath("PostgreSQLFunctions.TestConnection.sql"), "r") as file:
			commandText = file.read()
		exception = None
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
					result = cur.fetchone()[0]
					if (result):
						returnValue = json.loads(result)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		return returnValue

	@staticmethod
	def __GetDataColumn__(name:str, clrDataType:str):
		dataColumn:DataColumn = DataColumn(name)
		dataColumn = System.Type.GetType(clrDataType)
		return dataColumn

	@staticmethod
	def __GetCLRDataType__(obj):
		if (isinstance(obj, bool)):
			return "System.Boolean"
		elif (isinstance(obj, datetime.datetime)):
			return "System.DateTime"
		elif (isinstance(obj, int)):
			if (obj > 2147483647):
				return "System.Int64"
			else:
				return "System.Int32"
		else:
			return "System.String"

	@staticmethod
	def __ConvertToCLRDataType__(obj, exceptedType:str = None):
		returnValue = None
		if (obj is None):
			returnValue = System.DBNull.Value
		else:
			if (exceptedType is None):
				exceptedType = BulkData.__GetCLRDataType__(obj)
			match exceptedType:
				case "System.Boolean":
					returnValue = System.Convert.ToBoolean(obj)
				case "System.DateTime":
					returnValue = System.Convert.ToDateTime(obj.isoformat())
				case "System.Int32":
					returnValue = System.Convert.ToInt32(obj)
				case "System.Int64":
					returnValue = System.Convert.ToInt64(obj)
				case _:
					returnValue = System.Convert.ToString(obj)
		return returnValue

__all__ = ["PostgreSQLParameter", "PostgreSQLStoredProcedures", "PostgreSQLQueries", "PostgreSQLBulkData", "PostgreSQLClearData"]



import pandas
import datetime
import platform
import pythonnet
if (platform.system() == "Linux"):
	pythonnet.load("mono")
if (platform.system() == "Windows"):
	pythonnet.load("netfx")
import clr
clr.AddReference("System")
import System
from System.Data import DataTable
from System.Data import DataRow
from System.Data import DataColumn

class Statics:
	@staticmethod
	def GetCLRDataType(obj):
		if (isinstance(obj, bool)):
			return "System.Boolean"
		elif (isinstance(obj, datetime.datetime)):
			return "System.DateTime"
		elif (isinstance(obj, int)):
			if (obj > 2147483647):
				return "System.Int64"
			else:
				return "System.Int32"
		else:
			return "System.String"

	@staticmethod
	def ConvertToCLRDataType(obj, exceptedType:str = None):
		returnValue = None
		if (obj is None):
			returnValue = System.DBNull.Value
		else:
			if (exceptedType is None):
				exceptedType = Statics.GetCLRDataType(obj)
			match exceptedType:
				case "System.Boolean":
					returnValue = System.Convert.ToBoolean(obj)
				case "System.DateTime":
					returnValue = System.Convert.ToDateTime(obj.isoformat())
				case "System.Int32":
					returnValue = System.Convert.ToInt32(obj)
				case "System.Int64":
					returnValue = System.Convert.ToInt64(obj)
				case _:
					returnValue = System.Convert.ToString(obj)
		return returnValue

	@staticmethod
	def DataFrameToDataTable(dataFrame:pandas.DataFrame) -> DataTable:
		returnValue = DataTable()
		for index, row in dataFrame.iterrows():
			if (index == 0):
				for column in dataFrame.columns:
					dataColumn:DataColumn = DataColumn(column)
					dataColumn.DataType = System.Type.GetType(Statics.GetCLRDataType(row[column]))
					dataColumn.AllowDBNull  = True
					returnValue.Columns.Add(dataColumn)
			returnValue.Rows.Add(row.values.tolist())
		return returnValue

	@staticmethod
	def DataTableToDataFrame(dataTable:DataTable) -> pandas.DataFrame:
		returnValue:pandas.DataFrame = pandas.DataFrame()
		data:dict = dict()
		isFirstRow:bool = True
		for row in dataTable.Rows:
			if (isFirstRow):
				for column in dataTable.Columns:
					data.update({ column.ColumnName: list() })
				isFirstRow = False
			for index, column in enumerate(dataTable.Columns):
				data[column.ColumnName].append(row[index])
		returnValue = pandas.DataFrame(data)
		return returnValue

__all__ = ["Statics"]
